/* Creates TI-85 slideshows (compressed and uncompressed) from jpg,
     gif, bmp ... images
   Needs CxImage libraries! http://www.xdp.it/cximage.htm

Usage: mk85ss.exe [OPTIONS] [FILES]

  -e, --equalize
        Equalize images before quantisation
  -n, --normalize
        Normalize images before quantisation
  -s, --stretch
        Histogram stretch images before quantisation
  -d, --dither
        Dither images during quantisation
  -o <name>, --output <name>
        Save slideshow under <name>.85s (default: sshow.85s)

  -?, --help        Output this help

*/

#include <direct.h>
#include <afxwin.h>
#include "ximage.h"
#include "SimpleOpt.h"
#include "SimpleGlob.h"
#include "huffman.h"

void usage(char* argv[]) {
	printf("\nConvert Images to TI-85 slideshow format\n");
    printf("Usage: %s [OPTIONS] [FILES]\n",argv[0]);
	printf("\n");
	printf("  -e, --equalize\n");
	printf("        Equalize images before quantisation\n");
	printf("  -n, --normalize\n");
	printf("        Normalize images before quantisation\n");
	printf("  -s, --stretch\n");
	printf("        Histogram stretch images before quantisation\n");
	printf("  -d, --dither\n");
	printf("        Dither images during quantisation\n");
	printf("  -o <name>, --output <name>\n");
	printf("        Save slideshow under <name>.85s (uncompressed) and <name>c.85s\n");
	printf("        (compressed). <name> must not be longer than 7 characters\n");
	printf("        (default: sshow)\n");
	printf("\n");
	printf("  -?, --help        Output this help\n");
}

CSimpleOpt::SOption g_rgOptions[] = {
    { 1, "-e",		     SO_NONE    },
    { 1, "--equalize",   SO_NONE    },
    { 2, "-n",		     SO_NONE    },
    { 2, "--normalize",  SO_NONE    },
    { 3, "-s",		     SO_NONE    },
    { 3, "--stretch",    SO_NONE    },
    { 8, "-d",		     SO_NONE    },
    { 8, "--dither",	 SO_NONE    },
    { 9,"-o",		     SO_REQ_SEP },
    { 9,"--ouput",       SO_REQ_SEP },
    { 0, "-?",	         SO_NONE    },
    { 0, "--help",       SO_NONE    },
    SO_END_OF_OPTIONS
};

int main(int argc, char* argv[])
{
	int action=0; // do nothing
	char* fi=NULL;
	char* fo=NULL;
	int i,ii;
	unsigned long x,y;
	unsigned char *imgb;
	unsigned char *imgbnc;
	unsigned char *imgbc;
	bool dither=false;
	int cp=0,or=0;
	int cols=4;
	FILE *f=NULL;
	FILE *of=NULL;
    CSimpleGlob glob(SG_GLOB_NODOT|SG_GLOB_ONLYFILE);
	CSimpleOpt args(argc, argv, g_rgOptions);
	char *imgb2;
	char *imgbc2;

	// change into my directory
	char mydir[_MAX_PATH];
	char curdir[_MAX_PATH];
	GetModuleFileName(NULL,curdir,_MAX_PATH);
	strrchr(curdir,'\\')[0]=0;
	sprintf(mydir,"%s\\files",curdir);
	getcwd( curdir, _MAX_PATH );
	huff_decodenode_t nodes[MAX_TREE_NODES], *root, *node;
	huff_bitstream_t  stream;
	unsigned int      k, node_count;

	// parse command line
	fo="sshow";
    while (args.Next()) {
        if (args.LastError() == SO_SUCCESS) {
            if (args.OptionId() == 0) {
                usage(argv);
                return 0;
            }
			if (args.OptionId() == 9) {
				fo=args.OptionArg();
				if (strlen(fo)>7) {
					printf("%s is longer than 5 characters (use -h for help)\n", fo);
					return 1;
				}
			}
			else if (args.OptionId() == 8) {
				dither=true;}
			else {
				action = args.OptionId();}
        }
        else {
            printf("Invalid argument: %s (use -h for help)\n", args.OptionText());
            return 1;
        }
    }
    
	// glob my files
    if (glob.Add(args.FileCount(), args.Files()) != SG_SUCCESS ) {
        printf("Error while globbing files\n");
        return 1;
    }
	
	if (glob.FileCount() == 0) {usage(argv);return 1;}
	if (glob.FileCount() > 127) {
		printf("A slideshow can only contain up to 127 pictures\n");
		return 1;
	}

	// allocate my space
	imgbnc = (unsigned char *)malloc (2048*glob.FileCount());
	imgb = (unsigned char *)malloc (2048*glob.FileCount());
	imgbc = (unsigned char *)malloc (4096*glob.FileCount());
	imgb2=(char *)imgb+1024;
	imgbc2=(char *)imgbc+1024;

	for (ii=0;ii<glob.FileCount();ii++) {
		CxImage image;
		fi=glob.File(ii);

		printf ("Processing %s\n",fi);

		if (!image.Load(fi,0)){
	        fprintf(stdout, "%s\n", image.GetLastError());
	        fprintf(stdout, "error loading %s\n", fi);
	        return 1;
		}

		if (!image.GrayScale()) {
	        fprintf(stdout, "%s\n", image.GetLastError());
	        fprintf(stdout, "error grayscaling %s\n", fi);
	        return 1;
		}
	
		if (action==1 && !image.HistogramEqualize()) {
	        fprintf(stdout, "%s\n", image.GetLastError());
	        fprintf(stdout, "error equalizing %s\n", fi);
	        return 1;
		}
		if (action==2 && !image.HistogramNormalize()) {
	        fprintf(stdout, "%s\n", image.GetLastError());
	        fprintf(stdout, "error normalizing %s\n", fi);
	        return 1;
		}
		if (action==3 && !image.HistogramStretch()) {
	        fprintf(stdout, "%s\n", image.GetLastError());
	        fprintf(stdout, "error stretching %s\n", fi);
	        return 1;
		}
	
		if (image.GetWidth() < image.GetHeight() && !image.RotateLeft() ){
	        fprintf(stdout, "%s\n", image.GetLastError());
	        fprintf(stdout, "error rotating %s\n", fi);
	        return 1;
		}

		float f1=((float)128)/((float)image.GetWidth());
		float f2=((float)64)/((float)image.GetHeight());
		f1=f1<f2?f1:f2;

		long nw=(long)(image.GetWidth()*f1+.5);
		long nh=(long)(image.GetHeight()*f1+.5);
	
		if (!image.Resample2(nw,nh)) {
	        fprintf(stdout, "%s\n", image.GetLastError());
	        fprintf(stdout, "error resizing %s\n", fi);
	        return 1;
		}
	
		if (!image.DecreaseBpp(4,dither)) {
	        fprintf(stdout, "%s\n", image.GetLastError());
	        fprintf(stdout, "error downsampling %s\n", fi);
	        return 1;
		}

		//calculate TI-85 image
		int yo=(64-image.GetHeight())/2;
		int xo=(128-image.GetWidth())/2;
		for (i=0;i<2048;i++) {imgb[ii*2048+i]=0;imgbnc[ii*2048+i]=0;}

		for (y=image.GetHeight()-1;y!=-1;y--) {
			for (x=0;x<image.GetWidth();x++) {
				for (i=0;i<cols;i++) if (image.GetPixelGray(x,y)>=i*(256/cols) && image.GetPixelGray(x,y)<=(i+1)*(256/cols)) {cp=i;break;}
				or=1<<(7-(xo+x)%8);
				imgb[ii*2048+2*((63-yo-y)*16+(xo+x)/8)]|=(cp==2||cp==3)?0:or;		// img 1: black and dark  = 1
				imgb[ii*2048+1+2*((63-yo-y)*16+(xo+x)/8)]|=(cp==0||cp==3)?0:or;		// img 2: dark and light = 1 (for better compression)
				imgbnc[ii*2048+(63-yo-y)*16+(xo+x)/8]|=(cp==2||cp==3)?0:or;
				imgbnc[ii*2048+1024+(63-yo-y)*16+(xo+x)/8]|=(cp==1||cp==3)?0:or;
			}
		}
	}

	// ok, now we have all our images in imgb, save them

	chdir(mydir);
	// uncompressed
	if (!(of=fopen("tempo","wb"))) {
		fprintf(stdout, "error opening tempo\n");
	    goto ExitError;
	}
	fwrite(imgbnc,1,2048*glob.FileCount(),of);
	fclose(of);

	// now compress our data
	ii=Huffman_Compress(imgb,imgbc,2048*glob.FileCount());

	// decompress again to find our picture positions
	_Huffman_InitBitstream( &stream, imgbc );

	node_count = 0;
	imgbnc[0] = glob.FileCount();
	imgbnc[1] = (stream.BytePtr-imgbc+3+3*imgbnc[0]) % 256;
	imgbnc[2] = (stream.BytePtr-imgbc+3+3*imgbnc[0]) / 256;
	root = _Huffman_RecoverTree( nodes, &stream, &node_count );
	for (i=0; i<imgbnc[0]; i++) {
		imgbnc[3+i*3] = (stream.BytePtr-imgbc+3+3*imgbnc[0]) % 256;
		imgbnc[4+i*3] = (stream.BytePtr-imgbc+3+3*imgbnc[0]) / 256;
		imgbnc[5+i*3] = (1<<(7-stream.BitPos));
		for( k = 0; k < 2048; ++ k ) {
			node = root;
			while( node->Symbol < 0 ) {
				if( _Huffman_ReadBit( &stream ) ) node = node->ChildB;
				else node = node->ChildA;
			}
		}
	}

	// save the compressed pic
	if (!(of=fopen("tempoc","wb"))) {
		fprintf(stdout, "error opening tempoc\n");
		goto ExitError;
	}
	fwrite(imgbnc,1,3+imgbnc[0]*3,of);
	fwrite(imgbc,1,ii,of);
	fclose(of);

	// compile the slideshows (huffman and non-huffman)

	sprintf((char *)imgb,"%s\\%s.85s",curdir,fo);	// what are my output files?
	sprintf((char *)imgbc,"%s\\%sc.85s",curdir,fo);
	remove ((char *)imgb);
	remove ((char *)imgbc);

	sprintf((char *)imgb2,"c.bat sshow %s",fo);	// what are my temporary output files?
	sprintf((char *)imgbc2,"c.bat sshowc %sc",fo);
	system(imgb2);
	system(imgbc2);
	remove("tempo");
	remove("tempoc");
	sprintf((char *)imgb2,"%s.85s",fo);	// what are my temporary output files?
	sprintf((char *)imgbc2,"%sc.85s",fo);
	rename((char *)imgb2,(char *)imgb);
	rename((char *)imgbc2,(char *)imgbc);

	free (imgb);
	free (imgbnc);
	free (imgbc);
	printf("Done!\n");
	chdir(curdir);
	return 0;

ExitError:
	chdir(curdir);
	return 1;
}

